from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Set
from instaui.common.jsonable import Jsonable
from instaui.runtime import get_slot_stacks, pop_slot
from instaui.runtime._app import get_app_slot
from instaui.vars.slot_prop import BindingSlotPropItem

if TYPE_CHECKING:
    from instaui.components.component import Component

_DEFAULT_SLOT_NAME = ":"


class SlotManager(Jsonable):
    def __init__(self) -> None:
        super().__init__()
        self._slots: Dict[str, Slot] = {}

    def get_slot(self, name: str) -> Slot:
        name = _DEFAULT_SLOT_NAME if name == "default" else name

        if name not in self._slots:
            self._slots[name] = Slot(name)

        return self._slots[name]

    @property
    def default(self):
        return self.get_slot(_DEFAULT_SLOT_NAME)

    def _to_json_dict(self):
        if (
            len(self._slots) == 1
            and _DEFAULT_SLOT_NAME in self._slots
            and (not self._slots[_DEFAULT_SLOT_NAME]._has_props_use())
        ):
            return self._slots[_DEFAULT_SLOT_NAME]._children

        return {name: slot._to_json_dict() for name, slot in self._slots.items()}

    def has_slot(self) -> bool:
        return len(self._slots) > 0


class Slot(Jsonable):
    def __init__(self, name: str) -> None:
        super().__init__()

        self._id: Optional[str] = None
        self._name = name
        self._children: List[Component] = []
        self._props_use_name: Set[str] = set()

    def _has_props_use(self):
        return len(self._props_use_name) > 0

    def props(self, name: str):
        if self._id is None:
            self._id = get_app_slot().generate_slot_id()

        self._props_use_name.add(name)
        return BindingSlotPropItem(self._id, name)

    def __enter__(self):
        get_slot_stacks().append(self)
        return self

    def __exit__(self, *_):
        pop_slot()

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._children:
            data["items"] = self._children

        if self._props_use_name:
            data["props"] = {"id": self._id, "use": list(self._props_use_name)}

        return data



from abc import ABC, abstractmethod


class CanInputMixin(ABC):
    @abstractmethod
    def _to_input_config(self):
        pass


class CanOutputMixin(ABC):
    @abstractmethod
    def _to_output_config(self):
        pass



from .span import Span as span
from .label import Label as label
from .paragraph import Paragraph as paragraph
from .input import Input as input
from .number import Number as number
from .button import Button as button
from .checkbox import Checkbox as checkbox
from .form import Form as form
from .select import Select as select
from .ul import Ul as ul
from .li import Li as li
from .div import Div as div
from .range import Range as range
from .date import Date as date
from .link import Link as link

option = select.Option

__all__ = [
    "span",
    "label",
    "paragraph",
    "input",
    "number",
    "button",
    "checkbox",
    "form",
    "select",
    "option",
    "ul",
    "li",
    "div",
    "range",
    "date",
    "link",
]



from __future__ import annotations
import typing

from instaui.common.jsonable import Jsonable

from instaui.vars.path_var import PathVar
from instaui.vars.mixin_types.var_type import VarMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.py_binding import CanInputMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper

from . import _utils


class JsComputed(
    Jsonable,
    PathVar,
    VarMixin,
    ObservableMixin,
    CanInputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
):
    VAR_TYPE = "jsComputed"
    BIND_TYPE = "computed"

    def __init__(
        self,
        *,
        inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
        code: str = "",
        async_init_value: typing.Optional[typing.Any] = None,
    ) -> None:
        self.code = code

        sid, id = _utils.register_var(self)
        self._sid = sid
        self._id = id

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )

        self._async_init_value = async_init_value

    def __to_binding_config(self):
        return {
            "type": self.BIND_TYPE,
            "id": self._id,
            "sid": self._sid,
        }

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> typing.Dict:
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_pathable_binding_config(self) -> typing.Dict:
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()

        data["sid"] = self._sid
        data["id"] = self._id
        data["type"] = self.VAR_TYPE

        if self._inputs:
            data["inputs"] = self._inputs

        if sum(self._is_slient_inputs) > 0:
            data["slient"] = self._is_slient_inputs

        if sum(self._is_data) > 0:
            data["data"] = self._is_data

        if self._async_init_value is not None:
            data["asyncInit"] = self._async_init_value

        return data


TJsComputed = JsComputed



from __future__ import annotations
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Union
from .dataclass import JsLink, VueAppUse, VueAppComponent


class HtmlResource:
    use_tailwind: Optional[bool] = None
    title: Optional[str] = None
    favicon: Optional[Path] = None

    def __init__(self) -> None:
        self._css_links: Dict[Union[str, Path], Any] = {}
        self._style_tags: List[str] = []
        self._js_links: List[JsLink] = []
        self._script_tags: List[str] = []
        self._vue_app_use: Set[VueAppUse] = set()
        self._vue_app_components: Set[VueAppComponent] = set()
        self._import_maps: Dict[str, str] = {}
        self._appConfig = "{}"

    def add_css_link(self, link: Union[str, Path]):
        self._css_links[link] = None

    def add_style_tag(self, content: str):
        self._style_tags.append(content)

    def add_js_link(
        self,
        link: Union[str, Path],
        *,
        attrs: Optional[Dict[str, Any]] = None,
        insert_before: int = -1,
    ):
        if insert_before == -1:
            self._js_links.append(JsLink(link, attrs or {}))
            return
        self._js_links.insert(insert_before, JsLink(link, attrs or {}))

    def add_script_tag(
        self, content: str, script_attrs: Optional[Dict[str, Any]] = None
    ):
        self._script_tags.append(content)

    def add_vue_app_use(self, name: str):
        self._vue_app_use.add(VueAppUse(name))

    def add_vue_app_component(self, name: str, url: str):
        self._vue_app_components.add(VueAppComponent(name, url))

    def add_import_map(self, name: str, link: str):
        self._import_maps[name] = link



from __future__ import annotations
from typing import (
    Dict,
    Generic,
    Optional,
    TypeVar,
    Union,
    overload,
)

from instaui.common.jsonable import Jsonable
from instaui.vars.path_var import PathVar

from .mixin_types.var_type import VarMixin
from .mixin_types.py_binding import CanInputMixin, CanOutputMixin
from .mixin_types.observable import ObservableMixin
from .mixin_types.element_binding import ElementBindingMixin
from .mixin_types.pathable import CanPathPropMixin
from .mixin_types.str_format_binding import StrFormatBindingMixin
from . import _utils


_T_Value = TypeVar("_T_Value")


class Ref(
    Jsonable,
    PathVar,
    VarMixin,
    ObservableMixin,
    CanInputMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin[_T_Value],
    Generic[_T_Value],
):
    VAR_TYPE = "ref"

    def __init__(self, value: Optional[_T_Value] = None) -> None:
        self.value = value  # type: ignore

        sid, id = _utils.register_var(self)
        self._sid = sid
        self._id = id
        self._debounced = None

    def debounced(self, secounds: float):
        self._debounced = secounds
        return self

    def __to_binding_config(self):
        return {
            "type": self.VAR_TYPE,
            "id": self._id,
            "sid": self._sid,
        }

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_output_config(self):
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["sid"] = self._sid
        data["id"] = self._id
        data["type"] = self.VAR_TYPE

        if self._debounced is not None:
            data["debounced"] = self._debounced

        return data


TRef = Ref


@overload
def ref(value: Ref[_T_Value]) -> Ref[_T_Value]: ...


@overload
def ref(value: Optional[_T_Value] = None) -> Ref[_T_Value]: ...


def ref(value: Union[Ref[_T_Value], _T_Value, None] = None):
    if isinstance(value, Ref):
        return value
    return Ref(value)



from __future__ import annotations
from typing import (
    TypeVar,
)
from instaui.components.element import Element

_T = TypeVar("_T")


class Row(Element):
    def __init__(
        self,
    ):
        super().__init__("div")
        self.style("display: flex; flex-direction: row;")


    def gap(self, value: str) -> Row:
        return self.style({"gap": value})


from pathlib import Path
from typing import Union
from instaui.runtime import new_app_slot, reset_app_slot
from contextlib import contextmanager
from .func import to_html, to_html_str


@contextmanager
def scope():
    token = new_app_slot("zero")
    yield Wrapper()
    reset_app_slot(token)


class Wrapper:
    def to_html(self, file: Union[str, Path]):
        return to_html(file)

    def to_html_str(self):
        return to_html_str()



from __future__ import annotations
from typing import Dict, cast
from instaui.components.component import Component
from instaui.runtime._app import new_scope
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.types import TMaybeRef


class VIf(Component):
    def __init__(self, on: TMaybeRef[bool]):
        super().__init__("vif")
        self._on = cast(ElementBindingMixin, on)
        self.__scope_manager = new_scope(append_to_app=False)
        self.__scope = None

    def __enter__(self):
        self.__scope = self.__scope_manager.__enter__()
        return super().__enter__()

    def __exit__(self, *_) -> None:
        self.__scope_manager.__exit__(*_)
        return super().__exit__(*_)

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["props"] = {
            "on": self._on._to_element_binding_config(),
        }
        props: Dict = data["props"]

        props["scope"] = self.__scope

        if self._slot_manager.has_slot():
            props["items"] = self._slot_manager

        data.pop("slots", None)

        return data



from __future__ import annotations
from contextlib import contextmanager
import inspect
import os
import multiprocessing
from pathlib import Path
from typing import Any, Optional, Set
import __main__

from fastapi import FastAPI
from fastapi import Request
from fastapi.staticfiles import StaticFiles

from fastapi.responses import HTMLResponse
import uvicorn
from uvicorn.supervisors import ChangeReload
import itertools

from instaui.html_tools import to_config_data
from instaui.launch_collector import get_launch_collector
from instaui.page_info import PageInfo

from instaui import consts
from instaui.runtime._app import get_app_slot, get_default_app_slot
from instaui.runtime.dataclass import JsLink, VueAppComponent
from instaui.template import web_template


from . import dependency_router
from . import event_router
from . import watch_router
from . import debug_mode_router
from .middlewares import RequestContextMiddleware
from ._uvicorn import UvicornServer
from . import resource
from instaui.version import __version__ as _INSTA_VERSION

APP_IMPORT_STRING = "instaui.fastapi_server.server:Server._instance.app"


INSTAUI_STATIC_URL = f"/_instaui_{_INSTA_VERSION}/static"

VUE_JS_HASH_LINK = f"{INSTAUI_STATIC_URL}/{consts.VUE_ES_JS_PATH.name}"
INSTAUI_JS_HASH_LINK = f"{INSTAUI_STATIC_URL}/{consts.APP_ES_JS_PATH.name}"
APP_CSS_LINK = f"{INSTAUI_STATIC_URL}/{consts.APP_CSS_PATH.name}"
TAILWIND_JS_HASH_LINK = f"{INSTAUI_STATIC_URL}/{consts.TAILWIND_JS_PATH.name}"
FAVICON_LINK = f"{INSTAUI_STATIC_URL}/{consts.FAVICON_PATH.name}"


class Server:
    _instance: Optional[Server] = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self.app = FastAPI()
        self.app.add_middleware(RequestContextMiddleware)
        dependency_router.create_router(self.app)
        event_router.create_router(self.app)
        watch_router.create_router(self.app)
        debug_mode_router.create_router(self.app)

        self.add_instaui_static(self.app)

        for page_info in get_launch_collector()._page_router.values():
            self.register_page(page_info)

        self._registered_static_routes: Set[str] = set()

    def register_page(self, info: PageInfo):
        is_async = inspect.iscoroutinefunction(info.func)

        self._remove_route(info.path)

        if is_async:

            @self.app.get(info.path)
            async def _(request: Request):
                self._update_page_info(request, info)
                with _execute_request_lifespans():
                    await info.func()
                html = self._to_web_html(
                    page_info=info,
                    request=request,
                )

                return HTMLResponse(html)

        else:

            @self.app.get(info.path)
            def _(request: Request):
                self._update_page_info(request, info)
                with _execute_request_lifespans():
                    info.func()
                html = self._to_web_html(
                    page_info=info,
                    request=request,
                )

                return HTMLResponse(html)

    def _to_web_html(
        self,
        *,
        page_info: PageInfo,
        request: Request,
    ):
        config_data = to_config_data()

        system_slot = get_app_slot()
        default_app_slot = get_default_app_slot()
        html_resource = system_slot._html_resource
        default_html_resource = default_app_slot._html_resource

        favicon_url = FAVICON_LINK
        if html_resource.favicon:
            favicon_url = resource.record_resource(html_resource.favicon)
        else:
            if default_html_resource.favicon:
                favicon_url = resource.record_resource(default_html_resource.favicon)

        model = web_template.WebTemplateModel(
            vue_js_link=VUE_JS_HASH_LINK,
            instaui_js_link=INSTAUI_JS_HASH_LINK,
            css_links=[
                APP_CSS_LINK,
            ],
            config_dict=config_data,
            favicon_url=favicon_url,
            title=html_resource.title
            or default_html_resource.title
            or consts.PAGE_TITLE,
        )

        if html_resource.use_tailwind is None:
            if default_html_resource.use_tailwind:
                model.js_links.append(JsLink(TAILWIND_JS_HASH_LINK))
        else:
            if html_resource.use_tailwind:
                model.js_links.append(JsLink(TAILWIND_JS_HASH_LINK))

        # register custom components
        for component in system_slot._component_dependencies:
            if not component.esm:
                continue

            model.vue_app_component.append(
                VueAppComponent(
                    name=component.tag_name,
                    url=resource.record_resource(component.esm),
                )
            )

            if component.css:
                for css_link in component.css:
                    model.css_links.append(resource.record_resource(css_link))

        # register custom plugins
        for plugin in set(
            itertools.chain(
                system_slot._plugin_dependencies, default_app_slot._plugin_dependencies
            )
        ):
            if not plugin.esm:
                continue

            model.vue_app_use.append(plugin.name)

            model.add_extra_import_map(
                plugin.name, resource.record_resource(plugin.esm)
            )

            if plugin.css:
                for css_link in plugin.css:
                    model.css_links.append(resource.record_resource(css_link))

        # css file link to web static link
        for link, attrs in itertools.chain(
            html_resource._css_links.items(), default_html_resource._css_links.items()
        ):
            if isinstance(link, Path):
                model.css_links.append(resource.record_resource(link))

        # js file link to web static link
        for info in itertools.chain(
            html_resource._js_links, default_html_resource._js_links
        ):
            if isinstance(info.link, Path):
                model.js_links.append(JsLink((resource.record_resource(info.link))))

        for js_code in itertools.chain(
            html_resource._script_tags, default_html_resource._script_tags
        ):
            model.script_tags.append(js_code)

        for sylte_code in itertools.chain(
            html_resource._style_tags, default_html_resource._style_tags
        ):
            model.style_tags.append(sylte_code)

        model.prefix = request.headers.get(
            "X-Forwarded-Prefix", request.scope.get("root_path", "")
        )

        return web_template.render_web_html(model)

    def _update_page_info(self, request: Request, page_info: PageInfo):
        app = get_app_slot()

        app._page_path = page_info.path
        app._page_params = request.path_params
        app._query_params = dict(request.query_params)

    def _remove_route(self, path: str) -> None:
        self.app.routes[:] = [
            r for r in self.app.routes if getattr(r, "path", None) != path
        ]

    def try_close_server(self):
        UvicornServer.instance.should_exit = True

    def run(
        self,
        host="0.0.0.0",
        port=8080,
        reload: bool = True,
        reload_dirs: str = ".",
        reload_includes: str = "*.py",
        reload_excludes: str = ".*, .py[cod], .sw.*, ~*",
        log_level="info",
        workers: int | None = None,
        uds: str | None = None,
        **kwargs: Any,
    ):
        if multiprocessing.current_process().name != "MainProcess":
            return

        if reload and not hasattr(__main__, "__file__"):
            reload = False

        config = uvicorn.Config(
            APP_IMPORT_STRING if reload else self.app,
            host=host,
            port=port,
            reload=reload,
            log_level=log_level,
            workers=workers,
            uds=uds,
            reload_includes=_split_args(reload_includes) if reload else None,
            reload_excludes=_split_args(reload_excludes) if reload else None,
            reload_dirs=_split_args(reload_dirs) if reload else None,
            **kwargs,
        )

        UvicornServer.create_singleton(config, [debug_mode_router.when_server_reload])

        if config.should_reload:
            ChangeReload(config, target=UvicornServer.instance.run, sockets=[]).run()
        else:
            UvicornServer.instance.run()

        if config.uds:
            os.remove(config.uds)  # pragma: py-win32

    def run_with(self, app):
        assert isinstance(app, FastAPI), "app must be a FastAPI instance"

    @staticmethod
    def add_instaui_static(app: FastAPI):
        app.mount(
            INSTAUI_STATIC_URL,
            StaticFiles(directory=consts._STATIC_DIR),
            name=INSTAUI_STATIC_URL,
        )


def _split_args(args: str):
    return [a.strip() for a in args.split(",")]


@contextmanager
def _execute_request_lifespans():
    events = [iter(event()) for event in get_launch_collector().page_request_lifespans]
    for event in events:
        next(event)

    yield

    for event in events:
        try:
            next(event)
        except StopIteration:
            pass



from __future__ import annotations

from typing import TYPE_CHECKING, List
from instaui.common.jsonable import Jsonable


if TYPE_CHECKING:
    from instaui.vars.mixin_types.var_type import VarMixin
    from instaui.vars.mixin_types.py_binding import CanInputMixin
    from instaui.vars.web_computed import WebComputed
    from instaui.watch.web_watch import WebWatch
    from instaui.watch.js_watch import JsWatch
    from instaui.watch.vue_watch import VueWatch
    from instaui.vars.element_ref import ElementRef


class Scope(Jsonable):
    def __init__(self, id: str) -> None:
        super().__init__()
        self.id = id
        self._vars_id_counter = 0
        self._element_ref_id_counter = 0
        self._vars: List[VarMixin] = []
        self._web_computeds: List[WebComputed] = []
        self._element_refs: List[ElementRef] = []
        self._run_method_records: List = []
        self._web_watchs: List[WebWatch] = []
        self._js_watchs: List[JsWatch] = []
        self._vue_watchs: List[VueWatch] = []
        self._query = {}

    def set_run_method_record(
        self, scope_id: str, element_ref_id: str, method_name: str, args
    ):
        self._run_method_records.append((scope_id, element_ref_id, method_name, args))

    def generate_vars_id(self) -> str:
        self._vars_id_counter += 1
        return str(self._vars_id_counter)

    def generate_element_ref_id(self) -> str:
        self._element_ref_id_counter += 1
        return str(self._element_ref_id_counter)

    def register_element_ref(self, ref: ElementRef):
        self._element_refs.append(ref)

    def set_query(self, url: str, key: str, on: List[CanInputMixin]) -> None:
        self._query = {
            "url": url,
            "key": key,
            "on": [v._to_input_config() for v in on],
        }

    def register_var(self, var: VarMixin) -> None:
        self._vars.append(var)

    def register_web_computed(self, computed: WebComputed) -> None:
        self._web_computeds.append(computed)

    def register_web_watch(self, watch: WebWatch) -> None:
        self._web_watchs.append(watch)

    def register_js_watch(self, watch: JsWatch) -> None:
        self._js_watchs.append(watch)

    def register_vue_watch(self, watch: VueWatch) -> None:
        self._vue_watchs.append(watch)

    def _to_json_dict(self):
        data = super()._to_json_dict()
        if self._vars:
            data["vars"] = self._vars
        if self._query:
            data["query"] = self._query
        if self._web_watchs:
            data["py_watch"] = self._web_watchs
        if self._js_watchs:
            data["js_watch"] = self._js_watchs
        if self._vue_watchs:
            data["vue_watch"] = self._vue_watchs
        if self._element_refs:
            data["eRefs"] = self._element_refs
        if self._web_computeds:
            data["web_computed"] = self._web_computeds

        return data


class GlobalScope(Scope):
    def __init__(self, id: str) -> None:
        super().__init__(id)

    def register_var(self, var: VarMixin) -> None:
        raise ValueError("Can not register vars in global scope")

    def register_web_computed(self, computed: WebComputed) -> None:
        raise ValueError("Can not register web_computeds  in global scope")

    def register_web_watch(self, watch: WebWatch) -> None:
        raise ValueError("Can not register web_watchs  in global scope")

    def register_js_watch(self, watch: JsWatch) -> None:
        raise ValueError("Can not register js_watchs  in global scope")

    def register_vue_watch(self, watch: VueWatch) -> None:
        raise ValueError("Can not register vue_watchs  in global scope")



from __future__ import annotations
from typing import (
    Any,
    Dict,
)

from instaui.common.jsonable import Jsonable
from instaui.vars.path_var import PathVar

from .mixin_types.var_type import VarMixin
from .mixin_types.py_binding import CanInputMixin, CanOutputMixin
from .mixin_types.element_binding import ElementBindingMixin
from .mixin_types.pathable import CanPathPropMixin
from .mixin_types.str_format_binding import StrFormatBindingMixin
from . import _utils


class ConstData(
    Jsonable,
    PathVar,
    VarMixin,
    CanInputMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
):
    VAR_TYPE = "data"

    def __init__(self, value: Any = None) -> None:
        self.value = value  # type: ignore

        sid, id = _utils.register_var(self)
        self._sid = sid
        self._id = id

    def __to_binding_config(self):
        return {
            "type": self.VAR_TYPE,
            "id": self._id,
            "sid": self._sid,
        }

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_output_config(self):
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["sid"] = self._sid
        data["id"] = self._id
        data["type"] = self.VAR_TYPE

        return data


TConstData = ConstData



from fastapi import FastAPI
from fastapi.responses import FileResponse

from instaui.fastapi_server import resource


URL = f"{resource.URL}/{{hash_part:path}}"


def create_router(app: FastAPI):
    _dependency_handler(app)


def _dependency_handler(app: FastAPI):
    @app.get(URL)
    def _(hash_part: str) -> FileResponse:
        local_file = resource.get_by_hash(hash_part)

        return FileResponse(
            local_file, headers={"Cache-Control": "public, max-age=3600"}
        )



import threading
from typing import Callable, Dict, Optional

from instaui.runtime.context import get_context
from instaui.systems import func_system


ASYNC_URL = "/instaui/computed/async"
SYNC_URL = "/instaui/computed/sync"
INIT_URL = "/instaui/computed/init"

_computed_handlers: Dict[str, Callable] = {}
dict_lock = threading.Lock()


def register_handler(key: str, handler: Callable):
    if key in _computed_handlers:
        return
    with dict_lock:
        _computed_handlers[key] = handler


def get_handler(key: str) -> Optional[Callable]:
    return _computed_handlers.get(key)


def get_statistics_info():
    return {
        "_computed_handlers count": len(_computed_handlers),
        "_computed_handlers keys": list(_computed_handlers.keys()),
    }


def create_handler_key(
    page_path: str,
    handler: Callable,
):
    _, lineno, _ = func_system.get_function_location_info(handler)

    if get_context().debug_mode:
        return f"path:{page_path}|line:{lineno}"
    return f"{page_path}|{lineno}"



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Union,
    overload,
)
from typing_extensions import Self

from instaui.vars import Ref
from instaui.components.value_element import ValueElement
from instaui.components.element import Element
from instaui.vars.types import TMaybeRef
from instaui.components.vfor import VFor

if TYPE_CHECKING:
    import instaui.vars as ui_vars


_T_Select_Value = Union[List[str], str]


class Select(ValueElement[Union[List[str], str]]):
    def __init__(
        self,
        value: Union[_T_Select_Value, ui_vars.TMaybeRef[_T_Select_Value], None] = None,
    ):
        super().__init__("select", value, is_html_component=True)

    def vmodel(
        self,
        value: Ref[Any],
        *modifiers: Literal["lazy"],
    ):
        return super().vmodel(value, *modifiers)  # type: ignore

    @overload
    def on_change(self, handler: Callable, *, key: Optional[str] = None) -> Self: ...

    @overload
    def on_change(
        self,
        handler: str,
        *,
        bindings: Optional[Dict] = None,
        key: Optional[str] = None,
    ) -> Self: ...

    def on_change(
        self,
        handler: Union[Callable, str],
        *,
        bindings: Optional[Dict] = None,
        key: Optional[str] = None,
    ):
        self.on("change", handler, bindings=bindings, key=key)  # type: ignore
        return self

    @classmethod
    def from_list(
        cls,
        options: TMaybeRef[List],
    ) -> Select:
        with cls() as select:
            with VFor(options) as item:
                Select.Option(item)  # type: ignore

        return select

    class Option(Element):
        def __init__(
            self,
            text: Optional[ui_vars.TMaybeRef[str]] = None,
            value: Optional[ui_vars.TMaybeRef[str]] = None,
            disabled: Optional[ui_vars.TMaybeRef[bool]] = None,
        ):
            props = {
                key: value
                for key, value in {
                    "text": text,
                    "value": value,
                    "disabled": disabled,
                }.items()
                if value is not None
            }
            super().__init__("option")

            self._props.update(props)



from pathlib import Path
from typing import Dict
from instaui.systems import file_system
from instaui.version import __version__ as _INSTA_VERSION

URL = f"/_instaui_{_INSTA_VERSION}/resource"
_THashPart = str
_HASH_PART_MAP: Dict[_THashPart, Path] = {}
_FILE_URL_MAP: Dict[Path, _THashPart] = {}


def get_by_hash(hash_part: str) -> Path:
    return _HASH_PART_MAP[hash_part]


def record_resource(path: Path):
    if path in _FILE_URL_MAP:
        return _FILE_URL_MAP[path]

    hash_part = _generate_hash_part(path)
    _HASH_PART_MAP[hash_part] = path
    url = f"{URL}/{hash_part}"
    _FILE_URL_MAP[path] = url
    return url


def _generate_hash_part(path: Path):
    path = Path(path).resolve()
    is_file = path.is_file()

    if is_file:
        return f"{file_system.generate_hash_name_from_path(path.parent)}/{path.name}"

    return file_system.generate_hash_name_from_path(path)



from __future__ import annotations
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

_TTagName = str


@dataclass(frozen=True)
class ComponentDependencyInfo:
    tag_name: _TTagName = field(hash=True)
    esm: Path = field(hash=False)
    externals: Optional[List[Path]] = field(
        default_factory=list, compare=False, hash=False
    )
    css: List[Path] = field(default_factory=list, compare=False, hash=False)



from __future__ import annotations
from typing import Dict, Generic, Tuple, TypeVar, TYPE_CHECKING, Union, cast
from contextlib import contextmanager

from instaui.common.jsonable import Jsonable
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.path_var import PathVar

if TYPE_CHECKING:
    from instaui.components.vfor import VFor

_T = TypeVar("_T")


class VForItemProxy(
    PathVar,
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
    Jsonable,
    Generic[_T],
):
    def __init__(self, vfor_item: VForItem[_T]):
        self._vfor_item = vfor_item

    def __getattr__(self, name: str):
        return self._vfor_item[name]

    def __getitem__(self, name):
        return super().__getattribute__("_vfor_item")[name]

    def _to_element_binding_config(self) -> Dict:
        return super().__getattribute__("_vfor_item")._to_element_binding_config()

    def _to_input_config(self):
        return super().__getattribute__("_vfor_item")._to_input_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return super().__getattribute__("_vfor_item")._to_path_prop_binding_config()

    def _to_output_config(self):
        return super().__getattribute__("_vfor_item")._to_output_config()

    def _to_str_format_binding(self, order: int) -> Tuple[str, str]:
        return super().__getattribute__("_vfor_item")._to_str_format_binding(order)

    def _to_pathable_binding_config(self) -> Dict:
        return super().__getattribute__("_vfor_item")._to_pathable_binding_config()

    def _to_observable_config(self):
        return super().__getattribute__("_vfor_item")._to_observable_config()

    def _to_json_dict(self):
        return super().__getattribute__("_vfor_item")._to_json_dict()


class VForItem(
    PathVar,
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    CanPathPropMixin,
    ElementBindingMixin[_T],
    StrFormatBindingMixin,
    Generic[_T],
):
    VAR_Type = "vf"

    def __init__(self, vfor: VFor):
        super().__init__()
        self._vfor = vfor

    @property
    def dict_key(self):
        return self._vfor.current[1]

    @property
    def dict_value(self):
        return self._vfor.current[0]

    @property
    def proxy(self):
        return cast(_T, VForItemProxy(self))

    def _to_binding_config(self) -> Union[Jsonable, Dict]:
        return self._to_json_dict()

    def _to_element_binding_config(self):
        return self._to_json_dict()

    def _to_input_config(self):
        return self._to_json_dict()

    def _to_output_config(self):
        return self._to_json_dict()

    def _to_path_prop_binding_config(self) -> Dict:
        return self._to_json_dict()

    def _to_pathable_binding_config(self) -> Dict:
        return self._to_json_dict()

    def _to_observable_config(self):
        return self._to_json_dict()

    def _to_json_dict(self):
        data: Dict = {
            "type": self.VAR_Type,
            "fid": self._vfor._fid,
        }

        return data


class VForIndex(
    CanInputMixin,
    CanPathPropMixin,
    ElementBindingMixin,
    StrFormatBindingMixin,
):
    def __init__(self, vfor: VFor):
        super().__init__()
        self._vfor = vfor

    def _to_element_binding_config(self):
        return self._to_json_dict()

    def _to_input_config(self):
        return self._to_json_dict()

    def _to_path_prop_binding_config(self) -> Dict:
        return self._to_json_dict()

    def _to_json_dict(self):
        return {
            "type": "vf-i",
            "fid": self._vfor._fid,
        }


class VForDict(
    CanInputMixin,
    CanOutputMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
    Jsonable,
):
    def __init__(self, vfor: VFor):
        self._vfor = vfor

    @property
    def dict_key(self):
        return self._vfor.current[1]

    @property
    def dict_value(self):
        return self._vfor.current[0]

    @contextmanager
    def with_index(self):
        self.__enter__()
        yield self, cast(int, VForIndex(self._vfor))

    def __enter__(self):
        self._vfor.__enter__()
        return self

    def __exit__(self, *_) -> None:
        return self._vfor.__exit__(*_)

    def _to_element_binding_config(self) -> Dict:
        return self.dict_value._to_element_binding_config()

    def _to_input_config(self):
        return self.dict_value._to_input_config()

    def _to_output_config(self):
        return self.dict_value._to_output_config()

    def _to_json_dict(self):
        return self.dict_value._to_json_dict()


class VForWithIndex(Generic[_T]):
    def __init__(self, vfor: VFor[_T]):
        self._vfor = vfor

    def __enter__(self):
        self._vfor.__enter__()
        return cast(_T, self._vfor.current.proxy), cast(int, VForIndex(self._vfor))

    def __exit__(self, *_) -> None:
        return self._vfor.__exit__(*_)


TVForItem = VForItem
TVForIndex = VForIndex



from instaui.runtime._app import get_current_scope
from instaui.vars.mixin_types.var_type import VarMixin


def register_var(object: VarMixin):
    """
    sid,id = register_var(object)
    """
    scope = get_current_scope()
    scope.register_var(object)

    return scope.id, scope.generate_vars_id()



from typing import Dict
from instaui.common.jsonable import Jsonable
from instaui.vars.mixin_types.element_binding import ElementBindingMixin


class LambdaFunc(Jsonable, ElementBindingMixin):
    def __init__(
        self, code: str, *, bindings: Dict[str, ElementBindingMixin], computed=False
    ):
        self.code = code
        self.type = "js"
        self._bindings = bindings
        self._computed = computed

    def _to_binding_config(self) -> Dict:
        return self._to_json_dict()

    def _to_js_binding_config(self):
        return self._to_json_dict()

    def _to_element_binding_config(self):
        return self._to_json_dict()

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._bindings:
            data["bind"] = {
                k: v._to_element_binding_config() for k, v in self._bindings.items()
            }

        if self._computed is True:
            data["ext"] = ["cpt"]

        return data



from __future__ import annotations
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Union
from instaui.common.jsonable import dumps, dumps2dict
from instaui.runtime._app import get_app_slot


def add_css_link(href: Union[str, Path]):
    get_app_slot()._html_resource.add_css_link(href)


def add_js_link(
    link: Union[str, Path],
    *,
    type: Optional[Literal["module"]] = None,
):
    attrs = {
        "type": type,
    }

    get_app_slot()._html_resource.add_js_link(link, attrs=attrs)


def add_style(content: str):
    get_app_slot()._html_resource.add_style_tag(content)


def use_tailwind(value=True):
    get_app_slot()._html_resource.use_tailwind = value


def use_page_title(title: str):
    get_app_slot()._html_resource.title = title


def use_favicon(favicon: Path):
    get_app_slot()._html_resource.favicon = favicon


def add_js_code(code: str, *, script_attrs: Optional[Dict[str, Any]] = None):
    get_app_slot()._html_resource.add_script_tag(code, script_attrs=script_attrs)


def add_vue_app_use(name: str):
    get_app_slot()._html_resource.add_vue_app_use(name)


def to_config_data() -> Dict:
    return dumps2dict(get_app_slot())


def to_json(indent=False):
    return dumps(get_app_slot(), indent=indent)



from __future__ import annotations
import inspect
import typing
from typing_extensions import ParamSpec
from . import _types
from . import _utils

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_app_slot, get_current_scope
from instaui.handlers import watch_handler

from instaui.vars.mixin_types.py_binding import CanOutputMixin
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper

P = ParamSpec("P")
R = typing.TypeVar("R")


class WebWatch(Jsonable, typing.Generic[P, R]):
    def __init__(
        self,
        func: typing.Callable[P, R],
        inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
        outputs: typing.Optional[typing.Sequence[CanOutputMixin]] = None,
        immediate: bool = True,
        deep: typing.Union[bool, int] = True,
        once: bool = False,
        flush: typing.Optional[_types.TFlush] = None,
        _debug: typing.Optional[typing.Any] = None,
    ) -> None:
        get_current_scope().register_web_watch(self)

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )

        self._outputs = [output._to_output_config() for output in outputs or []]
        self._fn = func
        self._immediate = immediate
        self._deep = deep
        self._once = once
        self._flush = flush
        self._debug = _debug

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        app = get_app_slot()

        if app.mode == "web":
            hkey = watch_handler.create_handler_key(
                page_path=app.page_path,
                handler=self._fn,
            )

            watch_handler.register_handler(hkey, self._fn, len(self._outputs))

            url = (
                watch_handler.ASYNC_URL
                if inspect.iscoroutinefunction(self._fn)
                else watch_handler.SYNC_URL
            )

            data["url"] = url
            data["key"] = hkey
            if self._inputs:
                data["inputs"] = self._inputs

            if sum(self._is_slient_inputs) > 0:
                data["slient"] = self._is_slient_inputs

            if sum(self._is_data) > 0:
                data["data"] = self._is_data

            if self._debug:
                data["debug"] = self._debug

            if self._outputs:
                data["outputs"] = self._outputs

            if self._immediate is not True:
                data["immediate"] = self._immediate

            if self._deep is not True:
                _utils.assert_deep(self._deep)
                data["deep"] = self._deep
            if self._once is not False:
                data["once"] = self._once
            if self._flush is not None:
                data["flush"] = self._flush

            return data

        return {}


def watch(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    immediate: bool = True,
    deep: typing.Union[bool, int] = True,
    once: bool = False,
    flush: typing.Optional[_types.TFlush] = None,
    _debug: typing.Optional[typing.Any] = None,
):
    def wrapper(func: typing.Callable[P, R]):
        return WebWatch(
            func,
            inputs,
            outputs=outputs,
            immediate=immediate,
            deep=deep,
            once=once,
            flush=flush,
            _debug=_debug,
        )

    return wrapper



import typing
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.common.jsonable import Jsonable
from .event_mixin import EventMixin


class JsEvent(Jsonable, EventMixin):
    def __init__(
        self,
        code: str,
        inputs: typing.Optional[typing.Sequence[CanInputMixin]] = None,
        outputs: typing.Optional[typing.Sequence[CanOutputMixin]] = None,
    ):
        self._is_const_data = [
            int(not isinstance(input, CanInputMixin)) for input in inputs or []
        ]
        self._org_inputs = list(inputs or [])
        self._org_outputs = list(outputs or [])
        self._inputs = [
            input._to_input_config() if isinstance(input, CanInputMixin) else input
            for input in inputs or []
        ]
        self._outputs = [output._to_output_config() for output in outputs or []]
        self.code = code

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._inputs:
            data["inputs"] = self._inputs

        if self._outputs:
            data["set"] = self._outputs

        if sum(self._is_const_data) > 0:
            data["data"] = self._is_const_data

        return data

    def copy_with_extends(self, extends: typing.Sequence):
        return js_event(
            code=self.code,
            inputs=self._org_inputs + list(extends),
            outputs=self._org_outputs,
        )

    def event_type(self):
        return "js"


def js_event(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    code: str,
):
    return JsEvent(inputs=inputs, outputs=outputs, code=code)



from __future__ import annotations
from typing import (
    Any,
    Dict,
    Optional,
    Sequence,
    Union,
    cast,
)

from instaui.vars.mixin_types.observable import ObservableMixin

from . import _types
from . import _utils

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_current_scope


class VueWatch(Jsonable):
    def __init__(
        self,
        sources: Union[Any, Sequence],
        callback: str,
        *,
        bindings: Optional[Dict[str, Any]] = None,
        immediate: bool = False,
        deep: Union[bool, int] = False,
        once: bool = False,
        flush: Optional[_types.TFlush] = None,
    ) -> None:
        get_current_scope().register_vue_watch(self)

        self.code = callback

        if isinstance(sources, Sequence):
            self.on = [
                cast(ObservableMixin, varObj)._to_observable_config()
                for varObj in sources
            ]
        else:
            self.on = cast(ObservableMixin, sources)._to_observable_config()

        if bindings:
            self.bind = {
                k: cast(ObservableMixin, v)._to_observable_config()
                for k, v in bindings.items()
            }

        if immediate is not False:
            self.immediate = immediate

        if deep is not False:
            _utils.assert_deep(deep)
            self.deep = deep

        if once is not False:
            self.once = once

        if flush is not None:
            self.flush = flush



import typing
from . import _types
from . import _utils

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_current_scope

from instaui.vars.mixin_types.py_binding import CanOutputMixin
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper


class JsWatch(Jsonable):
    def __init__(
        self,
        code: str,
        inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
        outputs: typing.Optional[typing.Sequence[CanOutputMixin]] = None,
        immediate: bool = True,
        deep: typing.Union[bool, int] = False,
        once: bool = False,
        flush: typing.Optional[_types.TFlush] = None,
    ) -> None:
        get_current_scope().register_js_watch(self)

        self.code = code

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )
        self._outputs = [output._to_output_config() for output in outputs or []]

        if immediate is not True:
            self.immediate = immediate

        if deep is not False:
            _utils.assert_deep(deep)
            self.deep = deep

        if once is not False:
            self.once = once

        if flush is not None:
            self.flush = flush

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._inputs:
            data["inputs"] = self._inputs

        if sum(self._is_slient_inputs) > 0:
            data["slient"] = self._is_slient_inputs

        if sum(self._is_data) > 0:
            data["data"] = self._is_data

        if self._outputs:
            data["outputs"] = self._outputs

        return data


def js_watch(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    code: str = "",
    immediate: bool = True,
    deep: typing.Union[bool, int] = False,
    once: bool = False,
    flush: typing.Optional[_types.TFlush] = None,
):
    return JsWatch(code, inputs, outputs, immediate, deep, once, flush)



from __future__ import annotations
from typing import Optional
from fastapi import Request
from contextvars import ContextVar, Token

current_request_ctx :ContextVar[Optional[Request]]= ContextVar("current_request", default=None)



def set_current_request(request: Request) :
    return current_request_ctx.set(request)

def reset_current_request(token: Token) -> None:
    current_request_ctx.reset(token)


def get_current_request() -> Request:
    return current_request_ctx.get()  # type: ignore




from typing import Any, Dict, List, Optional
from instaui.runtime._app import get_app_slot
from instaui.skip import is_skip_output
import pydantic


class ResponseData(pydantic.BaseModel):
    values: Optional[List[Any]] = None
    skips: Optional[List[int]] = None


def update_app_page_info(data: Dict):
    app = get_app_slot()

    page_info = data.get("page", {})
    app._page_path = page_info["path"]

    if "params" in page_info:
        app._page_params = page_info["params"]

    if "queryParams" in page_info:
        app._query_params = page_info["queryParams"]


def response_data(outputs_binding_count: int, result: Any):
    data = ResponseData()
    if outputs_binding_count > 0:
        if not isinstance(result, tuple):
            result = [result]

        result_infos = [(r, int(is_skip_output(r))) for r in result]

        if len(result_infos) == 1 and result_infos[0][1] == 1:
            return data

        data.values = [0 if info[1] == 1 else info[0] for info in result_infos]
        skips = [info[1] for info in result_infos]

        if sum(skips) > 0:
            data.skips = skips

    return data



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Union
from instaui.components.element import Element

if TYPE_CHECKING:
    import instaui.vars as ui_vars


class Label(Element):
    def __init__(
        self,
        text: Union[Any, ui_vars.TMaybeRef[Any], None] = None,
    ):
        super().__init__("label")

        if text is not None:
            self.props(
                {
                    "innerText": text,
                }
            )



from __future__ import annotations
from typing import (
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Union,
    Sequence,
    Generic,
    TypeVar,
    overload,
)
import pydantic

from instaui.components.component import Component
from instaui.vars.vfor_item import VForItem, VForDict, VForWithIndex
from instaui.runtime._app import get_app_slot, new_scope

from instaui.vars.mixin_types.element_binding import ElementBindingMixin

_T = TypeVar("_T")


class VFor(Component, Generic[_T]):
    def __init__(
        self,
        data: Union[Sequence[_T], ElementBindingMixin[List[_T]]],
        *,
        key: Union[Literal["item", "index"], str] = "index",
    ):
        """for loop component.

        Args:
            data (Union[Sequence[_T], ElementBindingMixin[List[_T]]]): data source.
            key (Union[Literal[&quot;item&quot;, &quot;index&quot;], str]]): key for each item. Defaults to 'index'.

        Examples:
        .. code-block:: python
            items = ui.state([1,2,3])

            with ui.vfor(items) as item:
                html.span(item)

            # object key
            items = ui.state([{"name": "Alice", "age": 20}, {"name": "Bob", "age": 30}])
            with ui.vfor(items, key=":item=>item.name") as item:
                html.span(item.name)
        """

        super().__init__("vfor")
        self._data = data
        self._key = key
        self._fid = get_app_slot().generate_vfor_id()
        self.__scope_manager = new_scope(append_to_app=False)
        self.__scope = None
        self._num = None
        self._transition_group_setting = None

    def __enter__(self) -> _T:
        self.__scope = self.__scope_manager.__enter__()
        super().__enter__()
        return VForItem(self).proxy # type: ignore

    def __exit__(self, *_) -> None:
        self.__scope_manager.__exit__(*_)
        return super().__exit__(*_)

    def _set_num(self, num):
        self._num = num

    def transition_group(self, name="fade", tag: Optional[str] = None):
        self._transition_group_setting = {"name": name, "tag": tag}
        return self

    @property
    def current(self):
        return VForItem(self)

    def with_index(self):
        return VForWithIndex(self)

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["props"] = {"fid": self._fid}

        props: Dict = data["props"]
        if self._key is not None and self._key != "index":
            props["fkey"] = self._key

        if self._data is not None:
            if isinstance(self._data, ElementBindingMixin):
                props["bArray"] = self._data._to_element_binding_config()
            else:
                props["array"] = self._data

        if self._num is not None:
            props["num"] = self._num

        if self._transition_group_setting is not None:
            props["tsGroup"] = {
                k: v for k, v in self._transition_group_setting.items() if v is not None
            }

        props["scope"] = self.__scope

        if self._slot_manager.has_slot():
            props["items"] = self._slot_manager

        data.pop("slots", None)

        return data

    @overload
    @classmethod
    def range(cls, end: int) -> VFor[int]: ...

    @overload
    @classmethod
    def range(cls, end: ElementBindingMixin[int]) -> VFor[int]: ...

    @classmethod
    def range(cls, end: Union[int, ElementBindingMixin[int]]) -> VFor[int]:
        obj = cls(None)  # type: ignore

        num = (  # type: ignore
            end._to_element_binding_config()
            if isinstance(end, ElementBindingMixin)
            else end
        )

        obj._set_num(num)

        return obj  # type: ignore

    @classmethod
    def from_dict(
        cls, data: Union[Mapping, pydantic.BaseModel, ElementBindingMixin[Dict]]
    ):
        return VForDict(VFor(data))  # type: ignore



from instaui.runtime.context import get_context


def create_server(
    debug: bool = True,
):
    from instaui.fastapi_server.server import Server

    context = get_context()
    context._app_mode = "web"
    context._debug_mode = debug

    return Server.get_instance()



from __future__ import annotations
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Sequence,
    TypeVar,
)
from typing_extensions import ParamSpec

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_app_slot, get_current_scope
from instaui.handlers import watch_handler

from instaui.vars.path_var import PathVar
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper

P = ParamSpec("P")
R = TypeVar("R")


class WebComputed(
    Jsonable,
    PathVar,
    CanInputMixin,
    ObservableMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin[R],
    Generic[P, R],
):
    VAR_TYPE = "webComputed"

    def __init__(
        self,
        func: Callable[P, R],
        inputs: Optional[Sequence[TObservableInput]] = None,
        extend_outputs: Optional[Sequence[CanOutputMixin]] = None,
        init_value: Optional[R] = None,
        evaluating: Optional[CanOutputMixin] = None,
    ) -> None:
        scope = get_current_scope()

        self._sid = scope.id
        self._id = scope.generate_vars_id()
        get_current_scope().register_web_computed(self)

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )
        self._outputs = [output._to_output_config() for output in extend_outputs or []]
        self._fn = func
        self._init_value = init_value
        self._evaluating = evaluating

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        app = get_app_slot()

        if app.mode == "web":
            hkey = watch_handler.create_handler_key(
                page_path=app.page_path,
                handler=self._fn,
            )

            watch_handler.register_handler(hkey, self._fn, len(self._outputs) + 1)

            url = (
                watch_handler.ASYNC_URL
                if inspect.iscoroutinefunction(self._fn)
                else watch_handler.SYNC_URL
            )

            data["id"] = self._id
            data["sid"] = self._sid
            data["type"] = self.VAR_TYPE

            if self._inputs:
                data["inputs"] = self._inputs

            if self._outputs:
                data["outputs"] = self._outputs

            if sum(self._is_slient_inputs) > 0:
                data["slient"] = self._is_slient_inputs

            if sum(self._is_data) > 0:
                data["data"] = self._is_data

            data["url"] = url
            data["key"] = hkey
            if self._init_value is not None:
                data["init"] = self._init_value

            if self._evaluating:
                data["running"] = self._evaluating._to_output_config()

            return data

        return {}

    def __to_binding_config(self):
        return {
            "type": self.VAR_TYPE,
            "id": self._id,
            "sid": self._sid,
        }

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()


def web_computed(
    *,
    inputs: Optional[Sequence] = None,
    extend_outputs: Optional[Sequence] = None,
    init_value: Optional[Any] = None,
    evaluating: Optional[Any] = None,
):
    def wrapper(func: Callable[P, R]):
        return WebComputed(
            func,
            inputs=inputs,
            extend_outputs=extend_outputs,
            init_value=init_value,
            evaluating=evaluating,
        )

    return wrapper


TComputed = WebComputed



from __future__ import annotations
from typing import (
    Generic,
    List,
    Optional,
    Union,
    TypeVar,
)
from instaui.components.element import Element
from instaui import consts

from instaui.vars.types import TMaybeRef
from instaui.vars.web_computed import WebComputed


_T = TypeVar("_T")


class ValueElement(Element, Generic[_T]):
    def __init__(
        self,
        tag: Optional[str] = None,
        value: Union[_T, TMaybeRef[_T], None] = None,
        is_html_component: bool = False,
        value_name: str = "value",
    ):
        super().__init__(tag)
        self.__is_html_component = is_html_component

        if value is not None:
            if isinstance(value, WebComputed):
                self.props({value_name:value})
            else:
                self.vmodel(value, prop_name=value_name)

    def vmodel(
        self,
        value,
        modifiers: Union[consts.TModifier, List[consts.TModifier], None] = None,
        *,
        prop_name: str = "value",
    ):
        return super().vmodel(
            value,
            modifiers,
            prop_name=prop_name,
            is_html_component=self.__is_html_component,
        )



from __future__ import annotations
from typing import (
    Literal,
    Optional,
    TypeVar,
    Union,
)
from instaui.vars.types import TMaybeRef
from instaui.vars.js_computed import JsComputed
from instaui.components.element import Element
from instaui.vars.mixin_types.observable import ObservableMixin

_T = TypeVar("_T")


class Grid(Element):
    def __init__(
        self,
        rows: Optional[TMaybeRef[Union[int, str]]] = None,
        columns: Optional[TMaybeRef[Union[int, str]]] = None,
    ):
        super().__init__("div")
        self.style("display: grid;")

        if rows is not None:
            if isinstance(rows, int):
                rows = f"repeat({rows}, 1fr)"

            if isinstance(rows, ObservableMixin):
                rows = _convert_to_repeat_computed(rows)

            self.style({"grid-template-rows": rows})

        if columns is not None:
            if isinstance(columns, int):
                columns = f"repeat({columns}, 1fr)"

            if isinstance(columns, ObservableMixin):
                columns = _convert_to_repeat_computed(columns)

            self.style({"grid-template-columns": columns})

    @classmethod
    def auto_columns(
        cls,
        *,
        min_width: TMaybeRef[str],
        mode: TMaybeRef[Literal["auto-fill", "auto-fit"]] = "auto-fit",
    ) -> Grid:
        if isinstance(min_width, ObservableMixin) or isinstance(mode, ObservableMixin):
            template = JsComputed(
                inputs=[min_width, mode],
                code=r"(min_width, mode)=> `repeat(${mode}, minmax(min(${min_width},100%), 1fr))`",
            )

        else:
            template = f"repeat({mode}, minmax(min({min_width},100%), 1fr))"

        return cls(columns=template)

    def row_gap(self, gap: TMaybeRef[str]) -> Grid:
        return self.style({"row-gap": gap})

    def column_gap(self, gap: TMaybeRef[str]) -> Grid:
        return self.style({"column-gap": gap})

    def gap(self, gap: TMaybeRef[str]) -> Grid:
        return self.row_gap(gap).column_gap(gap)


def _convert_to_repeat_computed(value: ObservableMixin):
    return JsComputed(
        inputs=[value],
        code=r"""(value)=> {
    if (typeof value === "number"){
        return `repeat(${value}, 1fr)`
    }
    return value                     
    }""",
    )



from typing import Callable, Dict

ASYNC_URL = "/instaui/config/async"
SYNC_URL = "/instaui/config/sync"
_handlers: Dict[str, Callable] = {}


def register_handler(key: str, handler: Callable):
    _handlers[key] = handler


def get_handler(key: str) -> Callable:
    return _handlers[key]



from __future__ import annotations
from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_current_scope
from instaui.vars.mixin_types.py_binding import CanOutputMixin


class ElementRef(Jsonable, CanOutputMixin):
    def __init__(self) -> None:
        self._id = None
        self.__sid = get_current_scope().id

    def _set_id(self, id: str):
        self._id = id

    def __to_binding_config(
        self,
    ):
        return {
            "type": "ele_ref",
            "id": self._id,
            "sid": self.__sid,
        }

    def _to_output_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["id"] = self._id
        data["sid"] = self.__sid

        return data

    def _to_element_config(self):
        return {"id": self._id, "sid": self.__sid}


def run_element_method(method_name: str, *args, **kwargs):
    return {
        "method": method_name,
        "args": args,
    }



from .ref import Ref
from .web_computed import WebComputed
from .vfor_item import VForItem
from .types import TRefOrComputed, TMaybeRef


__all__ = [
    "Ref",
    "WebComputed",
    "TRefOrComputed",
    "TMaybeRef",
    "VForItem",
]



import inspect
import typing
from typing_extensions import ParamSpec
from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_current_scope, get_app_slot
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.handlers import event_handler
from .event_mixin import EventMixin


P = ParamSpec("P")
R = typing.TypeVar("R")
_T_input = typing.TypeVar("_T_input")
_T_output = typing.TypeVar("_T_output")


class WebEvent(Jsonable, EventMixin, typing.Generic[P, R]):
    def __init__(
        self,
        fn: typing.Callable[P, R],
        inputs: typing.List[CanInputMixin],
        outputs: typing.List[CanOutputMixin],
    ):
        self._inputs = inputs
        self._outputs = outputs
        self._fn = fn

        scope = get_current_scope()
        self._sid = scope.id

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def copy_with_extends(self, extends: typing.Sequence[CanInputMixin]):
        return WebEvent(
            fn=self._fn,
            inputs=self._inputs + list(extends),
            outputs=self._outputs,
        )

    def event_type(self):
        return "web"

    def _to_json_dict(self):
        app = get_app_slot()

        hkey = event_handler.create_handler_key(
            page_path=app.page_path, handler=self._fn
        )

        handler_path = (
            event_handler.ASYNC_URL
            if inspect.iscoroutinefunction(self._fn)
            else event_handler.SYNC_URL
        )

        event_handler.register_event_handler(
            hkey, self._fn, self._outputs, self._inputs
        )

        data = {}
        data["url"] = handler_path
        data["hKey"] = hkey
        data["sid"] = self._sid

        if self._inputs:
            data["bind"] = [
                binding._to_input_config()
                if isinstance(binding, CanInputMixin)
                else binding
                for binding in self._inputs
            ]

        if self._outputs:
            data["set"] = [ref._to_output_config() for ref in self._outputs]

        return data


@typing.overload
def ui_event(fn: typing.Callable[P, R]) -> WebEvent[P, R]: ...


@typing.overload
def ui_event(
    *,
    inputs: typing.Optional[typing.Union[_T_input, typing.Sequence[_T_input]]] = None,
    outputs: typing.Optional[
        typing.Union[_T_output, typing.Sequence[_T_output]]
    ] = None,
) -> typing.Callable[[typing.Callable[P, R]], WebEvent[P, R]]: ...


def ui_event(
    fn: typing.Optional[typing.Callable[P, R]] = None, *, inputs=None, outputs=None
) -> typing.Union[
    WebEvent[P, R], typing.Callable[[typing.Callable[P, R]], WebEvent[P, R]]
]:
    inputs = [inputs] if isinstance(inputs, CanInputMixin) else inputs
    outputs = [outputs] if isinstance(outputs, CanOutputMixin) else outputs
    if fn is None:

        def wrapper(fn: typing.Callable[P, R]):
            return WebEvent(fn, inputs=inputs or [], outputs=outputs or [])

        return wrapper

    return WebEvent(fn, inputs=inputs or [], outputs=outputs or [])



from __future__ import annotations
from typing import (
    Dict,
)
from instaui.common.jsonable import Jsonable

from instaui.vars.path_var import PathVar

from .mixin_types.py_binding import CanInputMixin
from .mixin_types.element_binding import ElementBindingMixin
from .mixin_types.pathable import CanPathPropMixin


class BindingSlotPropItem(
    Jsonable,
    PathVar,
    ElementBindingMixin,
    CanInputMixin,
    CanPathPropMixin,
):
    def __init__(self, slot_id: str, name: str) -> None:
        super().__init__()
        self.name = name
        self._id = slot_id

    def _to_element_binding_config(self):
        return self._to_binding_config()

    def _to_input_config(self):
        return self._to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self._to_binding_config()

    def _to_pathable_binding_config(self) -> Dict:
        return self._to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["type"] = "sp"
        data["id"] = self._id

        return data

    def _to_binding_config(self) -> Dict:
        return self._to_json_dict()



from __future__ import annotations
from typing import Literal
from instaui.components.element import Element
from instaui.vars.types import TMaybeRef


class Column(Element):
    def __init__(self):
        super().__init__("div")
        self.style("display: flex; flex-direction: column;gap:var(--insta-column-gap)")

    def gap(self, value: TMaybeRef[str]) -> Column:
        return self.style({"gap": value})

    def align_items(
        self, value: TMaybeRef[Literal["start", "end", "center", "stretch", "revert"]]
    ) -> Column:
        return self.style({"align-items": value})



from __future__ import annotations

import typing
from instaui.components.component import Component
from instaui.runtime._app import new_scope

from instaui.vars.mixin_types.element_binding import ElementBindingMixin


class Match(Component):
    def __init__(self, on: ElementBindingMixin):
        super().__init__("match")
        self._on = on
        self._default_case = None

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["props"] = {
            "on": self._on._to_element_binding_config(),
        }
        props: typing.Dict = data["props"]

        props["case"] = [
            item
            for item in self._slot_manager.default._children
            if isinstance(item, Case)
        ]

        if self._default_case:
            props["default"] = self._default_case

        data.pop("slots", None)

        return data

    def case(self, value: typing.Any) -> Case:
        with self:
            case = Case(value)

        return case

    def default(self) -> DefaultCase:
        with self:
            self._default_case = DefaultCase()

        return self._default_case


class Case(Component):
    def __init__(self, value: typing.Any):
        super().__init__("case")
        self._value = value
        self.__scope_manager = new_scope(append_to_app=False)
        self.__scope = None

    def __enter__(self):
        self.__scope = self.__scope_manager.__enter__()
        return super().__enter__()

    def __exit__(self, *_) -> None:
        self.__scope_manager.__exit__(*_)
        return super().__exit__(*_)

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["props"] = {
            "value": self._value,
        }
        props = data["props"]

        props["scope"] = self.__scope

        if self._slot_manager.has_slot():
            props["items"] = self._slot_manager

        data.pop("slots", None)
        return data


class DefaultCase(Component):
    def __init__(self):
        super().__init__("default-case")

        self.__scope_manager = new_scope(append_to_app=False)
        self.__scope = None

    def __enter__(self):
        self.__scope = self.__scope_manager.__enter__()
        return super().__enter__()

    def __exit__(self, *_) -> None:
        self.__scope_manager.__exit__(*_)
        return super().__exit__(*_)

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["props"] = {}
        props = data["props"]

        props["scope"] = self.__scope

        if self._slot_manager.has_slot():
            props["items"] = self._slot_manager

        data.pop("slots", None)
        return data
